/*
 * QEMU Xen PV Machine
 *
 * Copyright (c) 2007 Red Hat
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "vl.h"
#include "xenfb.h"
#include <linux/input.h>

/* A convenient function for munging pixels between different depths */
#define BLT(SRC_T,DST_T,RLS,GLS,BLS,RRS,GRS,BRS,RM,GM,BM)               \
    for (line = y ; line < h ; line++) {                                \
        SRC_T *src = (SRC_T *)(xenfb->pixels + (line*xenfb->row_stride) + (x*xenfb->depth/8)); \
        DST_T *dst = (DST_T *)(ds->data + (line*ds->linesize) + (x*ds->depth/8)); \
        int col;                                                        \
        for (col = x ; col < w ; col++) {                               \
            *dst = (((*src >> RRS)&RM) << RLS) |                        \
                (((*src >> GRS)&GM) << GLS) |                           \
                (((*src >> GRS)&BM) << BLS);                            \
            src++;                                                      \
            dst++;                                                      \
        }                                                               \
    }


/* This copies data from the guest framebuffer region, into QEMU's copy
 * NB. QEMU's copy is stored in the pixel format of a) the local X server (SDL case)
 * or b) the current VNC client pixel format.
 */
static void xen_pvfb_guest_copy(struct xenfb *xenfb, int x, int y, int w, int h)
{
    DisplayState *ds = (DisplayState *)xenfb->user_data;
    int line;

    if (xenfb->depth == ds->depth) { /* Perfect match can use fast path */
        for (line = y ; line < (y+h) ; line++) {
            memcpy(ds->data + (line * ds->linesize) + (x*ds->depth/8),
                   xenfb->pixels + (line*xenfb->row_stride) + (x*xenfb->depth/8),
                   w * xenfb->depth/8);
        }
    } else { /* Mismatch requires slow pixel munging */
        if (xenfb->depth == 8) {
            /* 8 bit source == r:3 g:3 b:2 */
            if (ds->depth == 16) {
                BLT(uint8_t, uint16_t,   5, 2, 0,   11, 5, 0,   7, 7, 3);
            } else if (ds->depth == 32) {
                BLT(uint8_t, uint32_t,   5, 2, 0,   16, 8, 0,   7, 7, 3);
            }
        } else if (xenfb->depth == 16) {
            /* 16 bit source == r:5 g:6 b:5 */
            if (ds->depth == 8) {
                BLT(uint16_t, uint8_t,    11, 5, 0,   5, 2, 0,    31, 63, 31);
            } else if (ds->depth == 32) {
                BLT(uint16_t, uint32_t,   11, 5, 0,   16, 8, 0,   31, 63, 31);
            }
        } else if (xenfb->depth == 32) {
            /* 32 bit source == r:8 g:8 b:8 (padding:8) */
            if (ds->depth == 8) {
                BLT(uint32_t, uint8_t,    16, 8, 0,   5, 2, 0,    255, 255, 255);
            } else if (ds->depth == 16) {
                BLT(uint32_t, uint16_t,   16, 8, 0,   11, 5, 0,   255, 255, 255);
            }
        }
    }
    dpy_update(ds, x, y, w, h);
}


/* Send a keypress from the client to the guest OS */
static void xen_pvfb_put_keycode(void *opaque, int keycode)
{
    struct xenfb *xenfb = (struct xenfb*)opaque;
    xenfb_send_key(xenfb, keycode & 0x80 ? 0 : 1, keycode & 0x7f);
}

/* Send a mouse event from the client to the guest OS */
static void xen_pvfb_mouse_event(void *opaque,
                                 int dx, int dy, int dz, int button_state)
{
    static int old_state = 0;
    int i;
    struct xenfb *xenfb = (struct xenfb*)opaque;
    DisplayState *ds = (DisplayState *)xenfb->user_data;
    if (xenfb->abs_pointer_wanted)
        xenfb_send_position(xenfb,
                            dx*ds->width/0x7fff,
                            dy*ds->height/0x7fff);
    else
        xenfb_send_motion(xenfb, dx, dy);

	for (i = 0 ; i < 8 ; i++) {
		int lastDown = old_state & (1 << i);
		int down = button_state & (1 << i);
		if (down == lastDown)
			continue;

		if (xenfb_send_key(xenfb, down, BTN_LEFT+i) < 0)
			return;
	}
    old_state = button_state;
}

/* QEMU display state changed, so refresh the framebuffer copy */
void xen_pvfb_update(void *opaque)
{
    struct xenfb *xenfb = (struct xenfb *)opaque;
    xen_pvfb_guest_copy(xenfb, 0, 0, xenfb->width, xenfb->height);
}

/* QEMU display state changed, so refresh the framebuffer copy */
void xen_pvfb_invalidate(void *opaque)
{
    struct xenfb *xenfb = (struct xenfb *)opaque;
    xen_pvfb_guest_copy(xenfb, 0, 0, xenfb->width, xenfb->height);
}

/* Screen dump is not used in Xen, so no need to impl this ? */
void xen_pvfb_screen_dump(void *opaque, const char *name) { }

void xen_pvfb_dispatch_store(void *opaque) {
    int ret;
    if ((ret = xenfb_dispatch_store((struct xenfb *)opaque)) < 0) {
        fprintf(stderr, "Failure while dispatching store: %d\n", ret);
        exit(1);
    }
}

void xen_pvfb_dispatch_channel(void *opaque) {
    int ret;
    if ((ret = xenfb_dispatch_channel((struct xenfb *)opaque)) < 0) {
        fprintf(stderr, "Failure while dispatching store: %d\n", ret);
        exit(1);
    }
}

/* The Xen PV machine currently provides
 *   - a virtual framebuffer
 *   - ....
 */
static void xen_init_pv(uint64_t ram_size, int vga_ram_size, char *boot_device,
			DisplayState *ds, const char **fd_filename,
			int snapshot,
			const char *kernel_filename,
			const char *kernel_cmdline,
			const char *initrd_filename, time_t timeoffset)
{
    struct xenfb *xenfb;
    extern int domid;
    int fd;

    /* Prepare PVFB state */
    xenfb = xenfb_new();
    if (xenfb == NULL) {
        fprintf(stderr, "Could not create framebuffer (%s)\n",
                strerror(errno));
        exit(1);
    }

    /* Talk to the guest */
    if (xenfb_attach_dom(xenfb, domid) < 0) {
        fprintf(stderr, "Could not connect to domain (%s)\n",
                strerror(errno));
        exit(1);
    }
    xenfb->update = xen_pvfb_guest_copy;
    xenfb->user_data = ds;

    /* Tell QEMU to allocate a graphical console */
    graphic_console_init(ds,
                         xen_pvfb_update,
                         xen_pvfb_invalidate,
                         xen_pvfb_screen_dump,
                         xenfb);

    /* Register our keyboard & mouse handlers */
    qemu_add_kbd_event_handler(xen_pvfb_put_keycode, xenfb);
    qemu_add_mouse_event_handler(xen_pvfb_mouse_event, xenfb,
                                 xenfb->abs_pointer_wanted
                                 );

    /* Listen for events from xenstore */
    fd = xenfb_get_store_fd(xenfb);
    if (qemu_set_fd_handler2(fd, NULL, xen_pvfb_dispatch_store, NULL, xenfb) < 0) {
        fprintf(stderr, "Could not register event handler (%s)\n",
                strerror(errno));
    }

    /* Listen for events from the event channel */
    fd = xenfb_get_channel_fd(xenfb);
    if (qemu_set_fd_handler2(fd, NULL, xen_pvfb_dispatch_channel, NULL, xenfb) < 0) {
        fprintf(stderr, "Could not register event handler (%s)\n",
                strerror(errno));
    }

    /* Setup QEMU display */
    dpy_resize(ds, xenfb->width, xenfb->height);
}

QEMUMachine xenpv_machine = {
    "xenpv",
    "Xen Para-virtualized PC",
    xen_init_pv,
};

/*
 * Local variables:
 *  indent-tabs-mode: nil
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
